""" Root module of package """

from .gacf import find_correlation_from_file, find_correlation_from_lists, \
    find_correlation_from_lists_cpp, GACF_LOG_MESSAGE
from .datastructure import EmptyDataStructureException, BadDataFileReadException



from .datastructure import DataStructure
from .correlator import CorrelationIterator, Correlator

SELECTION_FUNCTIONS = {
    'fast': Correlator.fastSelectionFunctionIdx,
    'natural': Correlator.naturalSelectionFunctionIdx
}

WEIGHT_FUNCTIONS = {
    'gaussian': Correlator.getGaussianWeights,
    'fractional': Correlator.getFractionWeights
}

GACF_LOG_MESSAGE = ' ######      ###     ######  ######## \n' \
                   '##    ##    ## ##   ##    ## ##       \n' \
                   '##         ##   ##  ##       ##       \n' \
                   '##   #### ##     ## ##       ######   \n' \
                   '##    ##  ######### ##       ##       \n' \
                   '##    ##  ##     ## ##    ## ##       \n' \
                   ' ######   ##     ##  ######  ##    \n' \
                   '------------------------------\n' \
                   'Number of Data Points: {no_data}\n' \
                   'Number of Lag Timesteps: {no_lag_points}\n' \
                   'Lag Resolution: {lag_resolution}\n' \
                   '------------------------------\n'


def set_up_correlation(corr, min_lag=None, max_lag=None, lag_resolution=None, alpha=None):
    """ No return type. Applies non-default values to correlator """
    if max_lag is not None:
        corr.max_lag = max_lag
    if min_lag is not None:
        corr.min_lag = min_lag
    if lag_resolution is not None:
        corr.lag_resolution = lag_resolution
    if alpha is not None:
        corr.alpha = alpha


def find_correlation(corr, selection_function='natural', weight_function='gaussian'):
    def lag_generator(min_lag, max_lag, lag_resolution):
        k = min_lag
        is_positive = False
        while k <= max_lag:
            if k == 0:
                is_positive = True
            if k > 0 and not is_positive:
                is_positive = True
                yield 0
            yield k
            k += lag_resolution

    num_steps = int(round((corr.max_lag - corr.min_lag) / corr.lag_resolution))

    for i, k in enumerate(lag_generator(corr.min_lag, corr.max_lag, corr.lag_resolution)):
        col_it = CorrelationIterator(k, corr.N_datasets)
        SELECTION_FUNCTIONS[selection_function](corr, col_it)
        corr.deltaT(col_it)
        WEIGHT_FUNCTIONS[weight_function](corr, col_it)
        corr.findCorrelation(col_it)
        corr.addCorrelationData(col_it, i)

    corr.cleanCorrelationData(i + 1)


def find_correlation_from_file(filename, min_lag=None, max_lag=None, lag_resolution=None, selection_function='natural',
                               weight_function='gaussian', alpha=None):
    """
    :param filename: string filename containing data in columns X, t, X_err
    :param max_lag: max lag in days
    :param lag_resolution: lag resolution in days
    :param selection_function: 'fast' or 'natural' - see paper for more details
    :param weight_function: 'gaussian' or 'fractional' see paper for more details
    :param alpha: weight function characteristic length scale, default is t.median_time
    :return: { 'lag_timeseries':[], 'correlations':[] }
    """
    ds = DataStructure(filename)
    corr = Correlator(ds)

    set_up_correlation(corr, min_lag, max_lag, lag_resolution, alpha)
    find_correlation(corr, selection_function, weight_function)
    # return a 3 tuple of lag timeseries, correlations, corr object
    return (corr.lag_timeseries(),
            corr.correlations() if len(corr.correlations()) > 1. else corr.correlations()[0],
            corr)


def find_correlation_from_lists(timeseries, values, errors=None, min_lag=None, max_lag=None, lag_resolution=None,
                                selection_function='natural', weight_function='gaussian', alpha=None):
    """
    :param timeseries: list of time values
    :param values: list of X values
    :param errors: (optional) list of errors on X values
    :param max_lag: max lag in days
    :param lag_resolution: lag resolution in days
    :param selection_function: 'fast' or 'natural' - see paper for more details
    :param weight_function: 'gaussian' or 'fractional' see paper for more details
    :param alpha: weight function characteristic length scale, default is t.median_time
    :return: { 'lag_timeseries':[], 'correlations':[] }
    """
    if errors is None:
        ds = DataStructure(timeseries, values)
    else:
        ds = DataStructure(timeseries, values, errors)
    corr = Correlator(ds)

    set_up_correlation(corr, min_lag, max_lag, lag_resolution, alpha)
    find_correlation(corr, selection_function, weight_function)
    return (corr.lag_timeseries(),
            corr.correlations() if len(corr.correlations()) > 1. else corr.correlations()[0],
            corr)


def find_correlation_from_lists_cpp(timeseries, values, errors=None, min_lag=None, max_lag=None, lag_resolution=None,
                                    alpha=None):
    """
        Using a fixed set up of natural selection function and linear weight function,
        will be faster than the python implementation in general.
       :param timeseries: list of time values
       :param values: list of X values
       :param errors: (optional) list of errors on X values
       :param max_lag: max lag in days
       :param lag_resolution: lag resolution in days
       :param alpha: weight function characteristic length scale, default is t.median_time
       :return: { 'lag_timeseries':[], 'correlations':[] }
       """
    if errors is None:
        ds = DataStructure(timeseries, values)
    else:
        ds = DataStructure(timeseries, values, errors)
    corr = Correlator(ds)
    set_up_correlation(corr, min_lag, max_lag, lag_resolution, alpha)
    corr.calculateStandardCorrelation()
    return (corr.lag_timeseries(),
            corr.correlations() if len(corr.correlations()) > 1. else corr.correlations()[0],
            corr)



import os
import re
import sys
import platform
import subprocess

from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext
from setuptools.command.test import test as TestCommand
from distutils.version import LooseVersion


class CMakeExtension(Extension):
    def __init__(self, name, sourcedir="", *args, **kwargs):
        Extension.__init__(self, name, sources=[], *args, **kwargs)
        self.sourcedir = os.path.abspath(sourcedir)


class PyTest(TestCommand):
    user_options = [("pytest-args=", "a", "Arguments to pass to py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = []

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import pytest

        errno = pytest.main(self.pytest_args)
        sys.exit(errno)


class CMakeBuild(build_ext):
    def run(self):
        try:
            out = subprocess.check_output(["cmake", "--version"])
        except OSError:
            raise RuntimeError(
                "CMake must be installed to build the following extensions: "
                + ", ".join(e.name for e in self.extensions)
            )

        if platform.system() == "Windows":
            cmake_version = LooseVersion(
                re.search(r"version\s*([\d.]+)", out.decode()).group(1)
            )
            if cmake_version < "3.1.0":
                raise RuntimeError("CMake >= 3.1.0 is required on Windows")

        for ext in self.extensions:
            self.build_extension(ext)

    def build_extension(self, ext):
        extdir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))
        cmake_args = [
            "-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=" + extdir,
            "-DPYTHON_EXECUTABLE=" + sys.executable,
        ]

        cfg = "Debug" if self.debug else "Release"
        build_args = ["--config", cfg]

        if platform.system() == "Windows":
            cmake_args += [
                "-DCMAKE_LIBRARY_OUTPUT_DIRECTORY_{}={}".format(cfg.upper(), extdir)
            ]
            if sys.maxsize > 2 ** 32:
                cmake_args += ["-A", "x64"]
            build_args += ["--", "/m"]
        else:
            try:
                GPP_COMPILER = (
                    subprocess.check_output("which g++", shell=True)
                    .decode()
                    .strip("\n\t\r")
                )
                GCC_COMPILER = (
                    subprocess.check_output("which gcc", shell=True)
                    .decode()
                    .strip("\n\t\r")
                )
            except subprocess.CalledProcessError as e:
                GPP_COMPILER, GCC_COMPILER = None, None
            else:
                cmake_args += [
                    "-DCMAKE_CXX_COMPILER=" + GPP_COMPILER,
                    "-DCMAKE_C_COMPILER=" + GCC_COMPILER,
                ]
            cmake_args += ["-DCMAKE_BUILD_TYPE=" + cfg]
            build_args += ["--", "-j2"]

        env = os.environ.copy()
        env["CXXFLAGS"] = '{} -DVERSION_INFO=\\"{}\\"'.format(
            env.get("CXXFLAGS", ""), self.distribution.get_version()
        )
        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)
        subprocess.check_call(
            ["cmake", ext.sourcedir] + cmake_args, cwd=self.build_temp, env=env
        )
        subprocess.check_call(
            ["cmake", "--build", "."] + build_args, cwd=self.build_temp
        )


setup(
    name="gacf",
    version="0.4.4",
    author="Joshua Briegal",
    author_email="jtb34@cam.ac.uk",
    description="A generalisation of the autocorrelation function, for non-uniformly sampled timeseries data",
    long_description="",
    packages=["gacf"],
    ext_modules=[
        CMakeExtension("gacf.correlator", "gacf"),
        CMakeExtension("gacf.datastructure", "gacf"),
    ],
    cmdclass=dict(build_ext=CMakeBuild, test=PyTest),
    tests_require=["pytest"],
    zip_safe=False,
)
